#include <stdio.h>
#include <gb/gb.h>
#include <gb/console.h>

#define GAME_BOARD_CELL_EMPTY     ' '
#define GAME_BOARD_CELL_PLAYER    'x'
#define GAME_BOARD_CELL_COMPUTER  'o'

#define GAME_STATUS_PLAYING   0
#define GAME_STATUS_WON       1
#define GAME_STATUS_LOST      2
#define GAME_STATUS_EQUALITY  3

UINT8 GAME_BOARD[9];  // (3×3 cases)
UINT8 GAME_CURSOR_X;
UINT8 GAME_CURSOR_Y;

void clear_screen(void) {
    UINT8 x;
    UINT8 y = 18;
    while (y) {
        y -= 1;
        x = 20;
        while (x) {
            x -= 1;
            gotoxy(x, y);
            setchar(' ');
        }
    }
}

void clear_line(UINT8 y) {
    UINT8 x = 20;
    while (x) {
        x -= 1;
        gotoxy(x, y);
        setchar(' ');
    }
}

UINT8 coord_2d_to_1d(UINT8 x, UINT8 y) {
    return y * 3 + x;
}

void title_screen(void) {
    // 1. Effacer tout ce qui se trouve à l'écran
    clear_screen();
    // 2. Afficher notre écran de titre
    gotoxy(4, 5);
    printf("Tic Tac Toe");
    gotoxy(3, 15);
    printf("- Press START -");
    // 3. Attendre que le joueur appuie sur START et qu'il relâche le bouton
    waitpad(J_START);
    waitpadup();
}

void game_init(void) {
    UINT8 i;

    for (i = 0 ; i < 9 ; i += 1) {
        GAME_BOARD[i] = GAME_BOARD_CELL_EMPTY;
    }

    // On initialise le curseur au centre du plateau de jeu
    GAME_CURSOR_X = 1;
    GAME_CURSOR_Y = 1;
}

void game_draw_board(void) {
    // 1. On efface l'écran
    clear_screen();
    // 2. On « dessine » le plateau
    gotoxy(4, 1);
    printf("Tic Tac Toe");
    gotoxy(1, 17);
    printf("X You - O Computer");
    gotoxy(0, 4);
    printf("       |   |   \n");
    printf("       |   |   \n");
    printf("       |   |   \n");
    printf("    ---+---+---\n");
    printf("       |   |   \n");
    printf("       |   |   \n");
    printf("       |   |   \n");
    printf("    ---+---+---\n");
    printf("       |   |   \n");
    printf("       |   |   \n");
    printf("       |   |   \n");
}

void game_draw_state(void) {
    UINT8 i;
    UINT8 x;
    UINT8 y;
    UINT8 graph_x;
    UINT8 graph_y;

    for (y = 0 ; y < 3 ; y += 1) {
        for (x = 0 ; x < 3 ; x += 1) {
            i = coord_2d_to_1d(x, y);
            graph_x = 4 + x * 4 + 1;
            graph_y = 4 + y * 4 + 1;
            gotoxy(graph_x, graph_y);
            setchar(GAME_BOARD[i]);
        }
    }

}

void game_draw_cursor(UINT8 cursor_char) {
    // Position du centre de la case sur l'écran
    UINT8 graph_x = 4 + GAME_CURSOR_X * 4 + 1;
    UINT8 graph_y = 4 + GAME_CURSOR_Y * 4 + 1;
    // Position du caractère à dessiner dans la case du plateau
    UINT8 cx;
    UINT8 cy;
    for (cy = graph_y - 1 ; cy <= graph_y + 1 ; cy += 1) {
        for (cx = graph_x - 1 ; cx <= graph_x + 1 ; cx += 1) {
            // On saute le caractère central (car il contient potentiellement
            // un pion, on ne veut donc pas le masquer
            if (cx == graph_x && cy == graph_y) {
                continue;
            }
            // On dessine le caractère à la position indiquée
            gotoxy(cx, cy);
            setchar(cursor_char);
        }
    }
}

void game_draw_game_over(UINT8 status) {
    clear_line(1);
    switch (status) {
        case GAME_STATUS_WON:
            gotoxy(4, 1);
            printf("YOU WON! :)");
            break;
        case GAME_STATUS_LOST:
            gotoxy(4, 1);
            printf("YOU LOST! :(");
            break;
        case GAME_STATUS_EQUALITY:
            gotoxy(5, 1);
            printf("EQUALITY!");
            break;
    }
    clear_line(17);
    gotoxy(2, 17);
    printf("- Press START -");
    waitpad(J_START);
}

void game_player_play(void) {
    UINT8 key;
    UINT8 i;

    // Boucle de jeu du joueur
    while (1) {
        // 1. Afficher le curseur
        game_draw_cursor('/');

        // 2. Attendre que le joueur appuie sur l'une des touches HAUT, BAS,
        // GAUCHE, DROITE, pour déplacer le curseur, ou sur la touche A pour
        // placer son pion.
        key = waitpad(J_UP | J_DOWN | J_LEFT | J_RIGHT | J_A);

        // 3. On efface le curseur
        game_draw_cursor(' ');

        // 4. On réagit en fonction du bouton appuyé par le joueur :

        //    * Si le joueur a appuyé sur une touche de direction, on déplace le
        //      curseur, en faisant attention à ne pas le déplacer en dehors des
        //      limites du plateau
        if (key & J_UP && GAME_CURSOR_Y != 0) {
            GAME_CURSOR_Y -= 1;
        }
        if (key & J_DOWN && GAME_CURSOR_Y != 2) {
            GAME_CURSOR_Y += 1;
        }
        if (key & J_LEFT && GAME_CURSOR_X != 0) {
            GAME_CURSOR_X -= 1;
        }
        if (key & J_RIGHT && GAME_CURSOR_X != 2) {
            GAME_CURSOR_X += 1;
        }

        //    * Si le joueur a appuyé sur A, on met son pion dans la case,
        //      si cette dernière est vide
        if (key & J_A) {
            i = coord_2d_to_1d(GAME_CURSOR_X, GAME_CURSOR_Y);
            if (GAME_BOARD[i] == GAME_BOARD_CELL_EMPTY) {
                GAME_BOARD[i] = GAME_BOARD_CELL_PLAYER;

                // 5. On met fin à la boucle de gameplay une fois le pion du
                // joueur placé.
                break;
            }
        }

        // On attend que le joueur ait relâché le bouton pour passer à la
        // suite, sinon il risque d'avoir du mal à sélectionner la case
        // qu'il souhaite ;)
        waitpadup();
    }
}

void game_computer_play(void) {
    UINT8 x;
    UINT8 y;
    UINT8 i;
    UINT8 player_score;
    UINT8 computer_score;
    UINT8 last_empty_cell;
    INT8 loose_cell = -1;

    // On parcourt les lignes
    for (y = 0 ; y < 3 ; y += 1) {
        player_score = 0;
        computer_score = 0;
        for (x = 0 ; x < 3 ; x += 1) {
            i = coord_2d_to_1d(x, y);
            switch (GAME_BOARD[i]) {
                case GAME_BOARD_CELL_EMPTY:
                    last_empty_cell = i;
                    break;
                case GAME_BOARD_CELL_PLAYER:
                    player_score += 1;
                    break;
                case GAME_BOARD_CELL_COMPUTER:
                    computer_score += 1;
                    break;
            }
        }
        // Si l'ordinateur a aligné 2 pions, on se place directement dans la
        // case vide pour remporter la partie
        if (computer_score == 2 && player_score == 0) {
            GAME_BOARD[last_empty_cell] = GAME_BOARD_CELL_COMPUTER;
            return;
        }
        // Si le joueur a aligné 2 pions, on enregistre la case qui pourrait
        // faire perdre l'ordinateur et on poursuit l'analyse
        if (player_score == 2 && computer_score == 0) {
            loose_cell = last_empty_cell;
        }
    }

    // On parcourt les colonnes
    for (x = 0 ; x < 3 ; x += 1) {
        player_score = 0;
        computer_score = 0;
        for (y = 0 ; y < 3 ; y += 1) {
            i = coord_2d_to_1d(x, y);
            switch (GAME_BOARD[i]) {
                case GAME_BOARD_CELL_EMPTY:
                    last_empty_cell = i;
                    break;
                case GAME_BOARD_CELL_PLAYER:
                    player_score += 1;
                    break;
                case GAME_BOARD_CELL_COMPUTER:
                    computer_score += 1;
                    break;
            }
        }
        if (computer_score == 2 && player_score == 0) {
            GAME_BOARD[last_empty_cell] = GAME_BOARD_CELL_COMPUTER;
            return;
        }
        if (player_score == 2 && computer_score == 0) {
            loose_cell = last_empty_cell;
        }
    }

    // On parcourt la première diagonale (case en haut à gauche jusqu'à la case
    // en bas à droite)
    player_score = 0;
    computer_score = 0;
    for (x = 0 ; x < 3 ; x += 1) {
        y = x;
        i = coord_2d_to_1d(x, y);
        switch (GAME_BOARD[i]) {
            case GAME_BOARD_CELL_EMPTY:
                last_empty_cell = i;
                break;
            case GAME_BOARD_CELL_PLAYER:
                player_score += 1;
                break;
            case GAME_BOARD_CELL_COMPUTER:
                computer_score += 1;
                break;
        }
    }
    if (computer_score == 2 && player_score == 0) {
        GAME_BOARD[last_empty_cell] = GAME_BOARD_CELL_COMPUTER;
        return;
    }
    if (player_score == 2 && computer_score == 0) {
        loose_cell = last_empty_cell;
    }

    // On parcourt la seconde diagonale (case en bas à gauche jusqu'à la case
    // en haut à droite)
    player_score = 0;
    computer_score = 0;
    for (x = 0 ; x < 3 ; x += 1) {
        y = 2 - x;
        i = coord_2d_to_1d(x, y);
        switch (GAME_BOARD[i]) {
            case GAME_BOARD_CELL_EMPTY:
                last_empty_cell = i;
                break;
            case GAME_BOARD_CELL_PLAYER:
                player_score += 1;
                break;
            case GAME_BOARD_CELL_COMPUTER:
                computer_score += 1;
                break;
        }
    }
    if (computer_score == 2 && player_score == 0) {
        GAME_BOARD[last_empty_cell] = GAME_BOARD_CELL_COMPUTER;
        return;
    }
    if (player_score == 2 && computer_score == 0) {
        loose_cell = last_empty_cell;
    }

    // Si on arrive jusqu'ici, c'est qu'il n'y a pas moyen immédiat de
    // gagner... On va donc vérifier s'il y a une possibilité pour le
    // joueur de gagner, et si c'est le cas, on va placer notre pion dans
    // la case qui lui permettrait de gagner.
    if (loose_cell != -1) {
        GAME_BOARD[loose_cell] = GAME_BOARD_CELL_COMPUTER;
        return;
    }

    // Sinon on place le pion dans la dernière case vide trouvée
    GAME_BOARD[last_empty_cell] = GAME_BOARD_CELL_COMPUTER;
}

UINT8 game_check_status(void) {
    UINT8 x;
    UINT8 y;
    UINT8 i;
    UINT8 player_score;
    UINT8 computer_score;
    UINT8 empty_cell = 0;

    // On parcourt les lignes
    for (y = 0 ; y < 3 ; y += 1) {
        // On (ré)initialise les scores
        player_score = 0;
        computer_score = 0;
        // On parcour chaque case de la ligne
        for (x = 0 ; x < 3 ; x += 1) {
            i = coord_2d_to_1d(x, y);
            // On regarde ce qui se trouve dans la case et on compte
            // les points
            switch (GAME_BOARD[i]) {
                case GAME_BOARD_CELL_EMPTY:
                    empty_cell = 1;
                    break;
                case GAME_BOARD_CELL_PLAYER:
                    player_score += 1;
                    break;
                case GAME_BOARD_CELL_COMPUTER:
                    computer_score += 1;
                    break;
            }
        }
        // Enfin, on regarde si quelqu'un a gagné
        if (player_score == 3) {
            return GAME_STATUS_WON;
        }
        if (computer_score == 3) {
            return GAME_STATUS_LOST;
        }
    }

    // On parcourt les colonnes
    for (x = 0 ; x < 3 ; x += 1) {
        player_score = 0;
        computer_score = 0;
        for (y = 0 ; y < 3 ; y += 1) {
            i = coord_2d_to_1d(x, y);
            switch (GAME_BOARD[i]) {
                // On ne revérifie pas pour les cases vides étant donné
                // que l'on a déjà parcouru tout le tableau dans la boucle
                // précédente...
                case GAME_BOARD_CELL_PLAYER:
                    player_score += 1;
                    break;
                case GAME_BOARD_CELL_COMPUTER:
                    computer_score += 1;
                    break;
            }
        }
        if (player_score == 3) {
            return GAME_STATUS_WON;
        }
        if (computer_score == 3) {
            return GAME_STATUS_LOST;
        }
    }

    // On parcourt la première diagonale (case en haut à gauche jusqu'à la case
    // en bas à droite)
    player_score = 0;
    computer_score = 0;
    for (x = 0 ; x < 3 ; x += 1) {
        y = x;
        i = coord_2d_to_1d(x, y);
        switch (GAME_BOARD[i]) {
            case GAME_BOARD_CELL_PLAYER:
                player_score += 1;
                break;
            case GAME_BOARD_CELL_COMPUTER:
                computer_score += 1;
                break;
        }
    }
    if (player_score == 3) {
        return GAME_STATUS_WON;
    }
    if (computer_score == 3) {
        return GAME_STATUS_LOST;
    }

    // On parcourt la seconde diagonale (case en bas à gauche jusqu'à la case
    // en haut à droite)
    player_score = 0;
    computer_score = 0;
    for (x = 0 ; x < 3 ; x += 1) {
        y = 2 - x;
        i = coord_2d_to_1d(x, y);
        switch (GAME_BOARD[i]) {
            case GAME_BOARD_CELL_PLAYER:
                player_score += 1;
                break;
            case GAME_BOARD_CELL_COMPUTER:
                computer_score += 1;
                break;
        }
    }
    if (player_score == 3) {
        return GAME_STATUS_WON;
    }
    if (computer_score == 3) {
        return GAME_STATUS_LOST;
    }

    if (empty_cell) {
        // Il reste au moins une case vide, on continue la partie
        return GAME_STATUS_PLAYING;
    } else {
        // Il n'y a plus de case vide, la partie est terminée
        return GAME_STATUS_EQUALITY;
    }
}

void game(void) {
    UINT8 status;

    game_init();
    game_draw_board();

    // Boucle principale de la partie
    while (1) {
        // Tour du joueur
        game_player_play();
        game_draw_state();
        status = game_check_status();
        if (status != GAME_STATUS_PLAYING) {
            break;  // on stoppe la boucle lorsque la partie est terminée
        }

        // Tour de l'ordinateur
        game_computer_play();
        game_draw_state();
        status = game_check_status();
        if (status != GAME_STATUS_PLAYING) {
            break;  // on stoppe la boucle lorsque la partie est terminée
        }
    }

    game_draw_game_over(status);
}

void main(void) {
    // Boucle principale du programme qui alternera à terme entre affichage de
    // l'écran de titre et la phase de jeu.
    while (1) {
        title_screen();
        game();
    }
}
